import json
import os
import logging
from datetime import datetime, timedelta, timezone
from typing import Optional
from dateutil.relativedelta import relativedelta
from time import sleep
from os import environ

from clients.beatsaver.client import Client as beatsaver_client
from clients.beatsaver.api.maps import get_maps_latest
from clients.beatsaver.models import MapDetail, GetMapsLatestSort

LOG_LEVEL = environ.get("LOG_LEVEL", "INFO")
logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=LOG_LEVEL
)

class BeatSaverAPI:
    BASE_URL = "https://api.beatsaver.com"

    def __init__(self, cache_expiry_days: int = 1, cache_dir: Optional[str] = None):
        self.client = beatsaver_client(base_url=self.BASE_URL)
        self.cache_expiry_days = cache_expiry_days
        self.CACHE_DIR = cache_dir or self._determine_cache_dir()
        if not os.path.exists(self.CACHE_DIR):
            os.makedirs(self.CACHE_DIR)
            logging.info(f"Created cache directory: {self.CACHE_DIR}")

    def _determine_cache_dir(self) -> str:
        home_cache = os.path.expanduser("~/.cache")
        saberlist_cache = os.path.join(home_cache, "saberlist")
        beatsaver_cache = os.path.join(saberlist_cache, "beatsaver")
        
        if os.path.exists(home_cache):
            if not os.path.exists(beatsaver_cache):
                try:
                    os.makedirs(beatsaver_cache)
                    logging.info(f"Created cache directory: {beatsaver_cache}")
                except OSError as e:
                    logging.warning(f"Failed to create {beatsaver_cache}: {e}")
                    return os.path.join(os.getcwd(), ".cache")
            return beatsaver_cache
        else:
            logging.info("~/.cache doesn't exist, using local .cache directory")
            return os.path.join(os.getcwd(), ".cache")

    def _get_cache_filename(self, map_id: str) -> str:
        return os.path.join(self.CACHE_DIR, f"map_{map_id}.json")

    def _is_cache_valid(self, cache_file: str) -> bool:
        if not os.path.exists(cache_file):
            return False
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(cache_file))
        return datetime.now() - file_modified_time < timedelta(days=self.cache_expiry_days)

    def get_maps(
            self, 
            use_cache: bool = True, 
            page_size: int = 20, 
            sort: GetMapsLatestSort = GetMapsLatestSort.FIRST_PUBLISHED, 
            verified: bool = True,
            month: Optional[int] = None,
            year: Optional[int] = None,
            max_pages: int = 50
    ) -> list[MapDetail]:
        """
        Fetches maps from BeatSaver for a specific month, handling pagination and caching.

        :param use_cache: Whether to use cached data if available.
        :param page_size: Number of maps per page.
        :param sort: Sorting criteria.
        :param verified: Whether to include maps made by verified mappers.
        :param month: The month to fetch maps for (1-12). Defaults to last month if None.
        :param year: The year to fetch maps for. Defaults to current year if None.
        :param max_pages: Maximum number of pages to fetch.
        :return: A list of MapDetail objects.
        """
        now = datetime.now(timezone.utc)
        if month is None or year is None:
            target_date = now - relativedelta(months=1)
            year = target_date.year
            month = target_date.month
        
        start_of_month = datetime(year, month, 1, tzinfo=timezone.utc)
        end_of_month = start_of_month + relativedelta(months=1)
        
        cache_file = os.path.join(self.CACHE_DIR, f"maps_{year}_{month:02d}_{sort.value}.json")

        cached_maps = []
        if use_cache and self._is_cache_valid(cache_file):
            logging.debug(f"Using cached data for maps (for {year}-{month:02d})")
            with open(cache_file, 'r') as f:
                cached_maps = [MapDetail.from_dict(map_data) for map_data in json.load(f)]

        if cached_maps:
            return cached_maps

        all_maps = cached_maps.copy()
        fetched_pages = 0
        last_map_date = end_of_month

        while fetched_pages < max_pages:
            maps = get_maps_latest.sync(
                client=self.client,
                page_size=page_size,
                sort=sort,
                verified=verified,
                automapper=False,
                before=last_map_date,
            )
            """Sample result:
            >>> maps
            SearchResponse(docs=[MapDetail(automapper=False, bl_qualified=False, bl_ranked=False, bookmarked=False, collaborators=[], created_at=datetime.datetime(2024, 10, 9, 21, 30, 16, 714066, tzinfo=tzutc()), curated_at=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, curator=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, declared_ai=<MapDetailDeclaredAi.NONE: 'None'>, deleted_at=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, description='', id='40c79', last_published_at=datetime.datetime(2024, 10, 9, 21, 30, 42, 709374, tzinfo=tzutc()), metadata=MapDetailMetadata(bpm=105.0, duration=153, level_author_name='PsychItsMike', song_author_name='TWRP', song_name='Content 4 U', song_sub_name='', additional_properties={}), name='TWRP - Content 4 U', qualified=False, ranked=False, stats=MapStats(downloads=0, downvotes=0, plays=0, reviews=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, score=0.5, score_one_dp=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, sentiment=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, upvotes=0, additional_properties={}), tags=[], updated_at=datetime.datetime(2024, 10, 9, 21, 30, 42, 709374, tzinfo=tzutc()), uploaded=datetime.datetime(2024, 10, 9, 21, 30, 42, 709373, tzinfo=tzutc()), uploader=UserDetail(admin=False, avatar='https://www.gravatar.com/avatar/5cff0b7698cc5a672c8544f0?d=retro', curator=False, curator_tab=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, description=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, email=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, follow_data=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, hash_='5cff0b7698cc5a672c8544f0', id=42848, name='psychitsmike', patreon=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, playlist_url='https://api.beatsaver.com/users/id/42848/playlist', senior_curator=False, stats=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, suspended_at=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, testplay=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, type=<UserDetailType.SIMPLE: 'SIMPLE'>, unique_set=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, upload_limit=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, verified_mapper=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, additional_properties={}), versions=[MapVersion(cover_url='https://na.cdn.beatsaver.com/c70447aa7c24e526bd2ef5ba9c647b833c5431d5.jpg', created_at=datetime.datetime(2024, 10, 9, 21, 30, 16, 714066, tzinfo=tzutc()), diffs=[MapDifficulty(bl_stars=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, bombs=5, characteristic=<MapDifficultyCharacteristic.STANDARD: 'Standard'>, chroma=False, cinema=False, difficulty=<MapDifficultyDifficulty.EXPERT: 'Expert'>, events=2382, label=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, length=255.0, max_score=507035, me=False, ne=False, njs=14.0, notes=559, nps=3.836, obstacles=10, offset=0.0, parity_summary=MapParitySummary(errors=0, resets=0, warns=0, additional_properties={}), seconds=145.714, stars=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, additional_properties={})], download_url='https://r2cdn.beatsaver.com/c70447aa7c24e526bd2ef5ba9c647b833c5431d5.zip', feedback=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, hash_='c70447aa7c24e526bd2ef5ba9c647b833c5431d5', key=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, preview_url='https://na.cdn.beatsaver.com/c70447aa7c24e526bd2ef5ba9c647b833c5431d5.mp3', sage_score=2, scheduled_at=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, state=<MapVersionState.PUBLISHED: 'Published'>, testplay_at=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, testplays=[], additional_properties={})], additional_properties={})], redirect=<clients.beatsaver.types.Unset object at 0x7fe4d3e8b810>, additional_properties={})
            """

            if not maps.docs:
                logging.warn("No more maps found.")
                break

            for map in maps.docs:
                logging.debug(
                    f"Fetched map: '{map.name}' by {map.metadata.song_author_name} [{map.uploader.name}, "
                #   f"created at {map.created_at.strftime('%Y-%m-%dT%H:%M:%S')}, "
                #   f"last published at {map.last_published_at.strftime('%Y-%m-%dT%H:%M:%S')}, "
                #   f"updated at {map.updated_at.strftime('%m-%d %H:%M')}, "
                    f"uploaded on {map.uploaded.strftime('%Y-%m-%dT%H:%M:%S')}"
                )

            new_maps = [map for map in maps.docs if map not in all_maps]
            logging.info(f"Fetched {len(new_maps)} new maps.")
            all_maps.extend(new_maps)

            if new_maps:
                last_map_date = new_maps[-1].updated_at.replace(tzinfo=timezone.utc)
            else:
                logging.info("No new maps in this batch, stopping.")
                break

            if last_map_date <= start_of_month:
                logging.info(f"Reached or passed the start of the month ({start_of_month}).")
                break

            fetched_pages += 1
            sleep(1)

        logging.debug(f"Total maps fetched: {len(all_maps)}")

        # Filter maps to ensure they're within the target month
        all_maps = [map for map in all_maps if start_of_month <= map.updated_at.replace(tzinfo=timezone.utc) <= end_of_month]
        logging.debug(f"Total maps after filtering: {len(all_maps)}")

        # Cache the results
        with open(cache_file, 'w') as f:
            json.dump([map.to_dict() for map in all_maps], f)

        logging.info(f"Cached {len(all_maps)} maps for {year}-{month:02d}")
        return all_maps


"""TESTING
from src.helpers.BeatSaverAPI import *
use_cache: bool = True
page_size: int = 20
sort: GetMapsLatestSort = GetMapsLatestSort.FIRST_PUBLISHED
verified: bool = True
month: Optional[int] = None
year: Optional[int] = None
max_pages: int = 2

import sys
# Check if we're in a REPL environment
def is_running_in_repl():
    return hasattr(sys, 'ps1') or sys.flags.interactive

# Create a module-level instance only if in a REPL
if is_running_in_repl():
    self = BeatSaverAPI()
    print("REPL environment detected. 'self' instance of BeatSaverAPI created for convenience.")
"""
